import psutil
import time
import tkinter as Tk
import plyer
from os.path import realpath
from _thread import start_new_thread


class Core:
    def __init__(self,core_id):
        self.data = psutil.sensors_temperatures
        self.core_id = core_id
        self.label = self.data()['coretemp'][core_id][0]
        self.high = self.data()['coretemp'][core_id][2]
        self.critical = self.data()['coretemp'][core_id][3]
    def current(self):
        return(self.data()['coretemp'][self.core_id][1])

def get_cores():
    core_objects = psutil.sensors_temperatures()['coretemp']
    cores = []
    [cores.append(Core(i)) for i in range(len(core_objects))]
    return(cores)

def print_core(core,show_percentages=True):
    high_percent = (core.current() / core.high) * 100
    critical_percent = (core.current() / core.critical) * 100
    print("[CORE]: {:<3} | [TEMP]: {} | [HIGH]: {} | [CRITICAL]: {}".format(core.label,core.current(),core.high,core.critical))
    if show_percentages == True:
        print("--- [{}%] of throttle heat ({}C)".format(round(high_percent,2),core.high))
        print("--- [{}%] of max heat ({}C)".format(round(critical_percent,2),core.critical))

class CoreWatch():
    def __init__(self,cores,check_frequency=10):
        self.running = False
        self.cores = cores
        self.app_name = "CoreWatch"
        self.check_frequency = check_frequency
        self.start()
    def _start_thread(self):
        print("CoreWatch started!")
        self.running = True
        while self.running == True:
            time.sleep(self.check_frequency) # pause between checks
            for core in self.cores:
                now = time.localtime(); timestring = "{:>2}:{:>2}:{:>2}".format(now[3],now[4],now[5])
                print("[{}] {} | core checked".format(timestring,core.label))
                if core.current() >= core.critical:
                    self._core_critical(core)
                elif core.current() >= core.high:
                    self._core_hot(core)
                elif core.current() > 0:
                    pass
                elif core.current() == 0:
                    self._core_abnormal(core)
                else:
                    self._core_abnormal(core)
            print("MEMORY USED: [{:<4}%] Virtual | [{:<4}%] Swap".format(psutil.virtual_memory()[2],psutil.swap_memory()[3]))
    def start(self):
        if self.running == True: print("CoreWatch already running!")
        else:
            start_new_thread(self._start_thread, ())
    def stop(self):
        if self.running == False: print("nothing is running!")
        else: self.running = False
    def _core_hot(self,hot_core):
        plyer.notification.notify(
            title="Core Hot",
            message="{} is running hot!".format(hot_core.label),
            app_name=self.app_name,
            app_icon=realpath("fire.ico"), # must be .ico on Windows
            timeout=10)
    def _core_critical(self,critical_core):
        plyer.notification.notify(
            title="Core Critical",
            message="{} at critical temperature!".format(critical_core.label),
            app_name=self.app_name,
            app_icon=realpath("exclaimation.png"),
            timeout=15)
    def _core_abnormal(self,abnormal_core):
        plyer.notification.notify(
            title="Core Abnormal",
            message="{} is acting abnormally".format(abnormal_core.label),
            app_name=self.app_name,
            app_icon=realpath("abnormal.png"),
            timeout=10)

# ---- SAMPLE CODE ---- #
#cores = get_cores()
#for core in cores: print_core(core,show_percentages=False)
# ---- END SAMPLE CODE ---- #

# class Window(Tk.Frame):
#     def __init__(self,master=None):
#         Tk.Frame.__init__(self,master)
#         self.master = master
#         self.i = 0
#         self._init_window()
#         self.poll_rate = 1000
#         self._updater()
#     def _init_window(self):
#         m = self.master
#         temps = Tk.Frame(m)
#         self.v = Tk.StringVar()
#         temp_text = Tk.Label(m,textvariable=self.v)
#         temp_text.pack()
#         temps.pack()
#     def update_temps(self):
#         self.i += 1
#         out = [self.i]
#         [out.append(core.current()) for core in cores]
#         self.v.set(out)
#     def _updater(self):
#         self.update_temps()
#         self.after(self.poll_rate,self._updater)
# root = Tk.Tk()
# root.geometry("400x400")
# app = Window(root)
# root.mainloop()



class BarMode:
    def __init__(self,name,color):
        self.name = name
        self.color = color
class VolumeBar:
    def __init__(self):
        self.modes = {
            'volume': BarMode('volume','GREEN'),
            'brightness': BarMode('brightness','BLUE'),
            'muted': BarMode('muted','RED'),
            'unknown': BarMode('unknown','GREY')}
        self.mode = self.modes['volume']



#TODO: fix volume bar not going to zero
#TODO: replace zero volume with mute if supported under Debian
#TODO: add animation options (e.g. slide in, fade in, etc.)
#TODO: add numbers to the bar
#TODO: add support for Windows/Mac (e.g. scroll support)
#TODO: allow bar to be attached to different sides
#TODO: allow customization of bar thickness
#TODO: make bar opacity adapt to darkness/brightness of background
#TODO: modifier key makes scroll change channel pan instead

# DEPENDENCIES:
# (all working with most recent versions as of 04/12/2018)
# sudo apt install python3-tk
# sudo apt install python3-dbus # maybe?
# sudo apt install libasound2-dev
# pip3 install pyalsaaudio --user
# pip3 install plyer --user
# pip3 install psutil --user

# BUILTIN MODULES #
import sys
import time # used for delays
import math
from _thread import start_new_thread # used to run functions in parallel
import tkinter as Tk
from subprocess import call
from os.path import realpath
# SITE PACKAGES #
from plyer import notification
import alsaaudio as al
# LOCAL MODULES
import volux.temperatures as temps
from volux.dp_datatools import LivePercentage, clamp
from volux.VolumeAssistant import VolumeAssistant
from volux.VolumeBar import VolumeBar
### ---- PREFERENCES ---- ###
program_title = "volux"
program_icon = realpath('icon.png')
sound_device = "Master"
default_mixer_name = "Master"
default_opacity = 0.5
outside_zone_opacity = 0.1
bar_height = 5
### ---- SETUP STUFF ---- ###
VolAs = VolumeAssistant() # initialise a Volume Assistant object
VolBar = VolumeBar()
coreWatch = temps.CoreWatch(temps.get_cores()) # start watching cores for temperature issues

### DEFINE STATES
class VolumeMode:
    def __init__(self):pass
    def enter(self):
        VolBar.mode = VolBar.modes['volume']
        VolAs.mixer.setmute(0)
    def vacate(self):
        VolBar.mode = VolBar.modes['unknown']
class MuteMode:
    def __init__(self): pass
    def enter(self):
        VolBar.mode = VolBar.modes['muted']
        VolAs.mixer.setmute(1)
    def vacate(self):
        VolBar.mode = VolBar.modes['unknown']
class BrightnessMode:
    def __init__(self): pass
    def enter(self):
        VolBar.mode = VolBar.modes['brightness']
    def vacate(self):
        VolBar.mode = VolBar.modes['unknown']
        if VolAs.ismuted() == True:
            return(MuteMode)
        elif VolAs.ismuted() == False:
            return(VolumeMode)
        else: raise TypeError("_ismuted should be a bool value")
### DEFINE STATE MANAGER
class StateManager:
    def __init__(self,initial_state):
        self.state = initial_state
    def change_state(self,new_state): # request to change states
        self.state().vacate()
        new_state().enter()
        self.state = new_state
### CREATE A STATE MANAGER
sm = StateManager(VolumeMode)

### ---- TKINTER STUFF BEGINS ---- ###
root = Tk.Tk()
class Window(Tk.Frame):
    def __init__(self,master=None):
        Tk.Frame.__init__(self,master)
        self.master = master
        self._init_objects()
        self._init_window()
        self._open_message()
    def _init_objects(self):
        self.displaySize = VolAs._get_display_size(root) # max size of the percentage bar in pixels
        self.barWidth = LivePercentage(0,self.displaySize['x']) # set width of bar
    def _init_window(self):
        m = self.master
        m.title("Please submit an issue to Github if you see this!")
        self.barHeight = bar_height # set height of bar            self._update_bar()
        barContainer = Tk.Frame(m)
        barContainer.configure(background="BLACK")
        barContainer.pack(fill=Tk.BOTH,expand=1)
        self.bar = Tk.Frame(barContainer) # create the bar
        self._update_bar() # update bar values
        def _adjust_bar(event,movement):
            if type(movement) == int: # if movement is an integer
                #self.barMode = self.barModes['volume']
                notchMultiplier = 5 # impact of a single scroll notch on percentage
                newVol = VolAs.volume + movement*notchMultiplier
                VolAs.volume = clamp(newVol,0,100)
            else: raise TypeError("Value should be an integer! Not sure what happened!")
            self._update_bar() # update the bar's graphical appearance
            self._update_volume() # update the system volume
            #TODO: support for Windows/Mac scrolling
        def _scroll_up(event):
            if sm.state == VolumeMode:
                _adjust_bar(event,+1)
            elif sm.state == BrightnessMode:
                _brightness_up()
            elif sm.state == MuteMode:
                sm.change_state(VolumeMode)
            self._update_bar()
        def _scroll_down(event):
            if sm.state == VolumeMode:
                _adjust_bar(event,-1)
            elif sm.state == BrightnessMode:
                _brightness_down()
            elif sm.state == MuteMode:
                sm.change_state(VolumeMode)
            self._update_bar()
        def _middle_click(event):
            if sm.state == VolumeMode: # if unmuted
                sm.change_state(MuteMode) # change to muted
                self._update_bar()
            elif sm.state == MuteMode: # if unmuted
                sm.change_state(VolumeMode) # change to muted
            self._update_bar()
        def _key_pressed(event):
            print("key pressed",event.key)
        def _key_released(event):
            print("key released",event.key)
        def _brightness_up(): print("WIP:"+"UP")
        def _brightness_down(): print("WIP:"+"DOWN")
        def _right_click(event):
            if sm.state == BrightnessMode:
                sm.change_state(sm.state().vacate())
            else:
                sm.change_state(BrightnessMode)
            self._update_bar()
            #print("brightness mode!")
        def _brightness_mode_off():
            sm.state_change(sm.state.vacate())
            self.barMode = self.barModes['default']
            self._update_bar()
            print("brightness mode off!")
        self.bar.pack(fill=Tk.Y,ipadx=5,ipady=5,side=Tk.LEFT)
        m.bind("<MouseWheel>",_adjust_bar)
        m.bind("<Button-2>",_middle_click)
        m.bind("<Button-4>",_scroll_up)
        m.bind("<Button-5>",_scroll_down)
        m.bind("<Button-3>",_right_click)
        m.bind("<Control-Button-4>",_brightness_up)
        m.bind("<Control-Button-5>",_brightness_down)
        m.bind("<Double-Button-3>",self._exit_app)
        barContainer.bind("<Enter>",self._mouse_entered)
        barContainer.bind("<Leave>",self._mouse_left)
    def _update_loop(self,ms_per_loop=1000):
        root.lift() # ensure window on top of others
        self._update_bar() # update bar graphics
        self.after(ms_per_loop,self._update_loop) # repeat _update_loop()
    def _update_bar(self):
        modeColor = VolBar.mode.color # set background based on mode color
        self.barWidth.setPerc(VolAs.volume) # set the width as a percentage
        newWidth = self.barWidth.getNum() # get a numerical version of the percentage
        self.bar.configure(background=modeColor,width=str(newWidth)) # update the bar with these settings
    def _update_volume(self):
        try: self.mixer.setvolume(VolAs.volume)
        except: call(["amixer","sset",str(VolAs.device),str(VolAs.volume)+"%","-q"])
    def _update_mute(self):
        muted = self.mixer.getmute()
        if muted[0] == True: self.mixer.setmute(0)
        elif muted[0] == False: self.mixer.setmute(1)
        else: raise Exception("mixer's .getmute()[0] method should return True or False!")
    def _mouse_entered(self,event): root.wm_attributes("-alpha",default_opacity)
    def _mouse_left(self,event): root.wm_attributes("-alpha",outside_zone_opacity)
    def _open_message(self):
        notification.notify(
            title=program_title,
            message="{} launched!".format(program_title),
            app_name=program_title,
            app_icon=program_icon,
            timeout=5)
    def _exit_app(self,event):
        notification.notify(
            title=program_title,
            message="{} now closing...".format(program_title),
            app_name=program_title,
            app_icon=program_icon,
            timeout=10)
        exit()
        
app = Window(root)
dispSize = VolAs._get_display_size(root)
overlay_w = dispSize['x']
overlay_h = app.barHeight
windowOffsets = {'x': 0,
                 'y': dispSize['y']-app.barHeight}
root.geometry("{}x{}+{}+{}".format(overlay_w,overlay_h,
                                   windowOffsets['x'],windowOffsets['y'])) # define the size of the window
root.attributes("-topmost",True) # force window to stay on top (doesn't work in full screen applications)
root.overrideredirect(1) # remove frame of window
root.wait_visibility(root) # required for window transparency
root.wm_attributes("-alpha",outside_zone_opacity) # make window transparent
root.title(program_title)

print(sys.argv[0])
if '__main__.py' in sys.argv[0]:
    app._update_loop() # must be before main loop
    root.mainloop()



class LivePercentage:
    def __init__(self,minimum,maximum):
        self.min = minimum # minimum number of range
        self.max = maximum # maximum number of range
        self.val = None
    def _percOf(self,number): return(((number - self.min) * 100) / (self.max - self.min))
    def _numOf(self,percentage): return(((percentage * (self.max - self.min) / 100) + self.min))
    def getPerc(self,*args): return(self._percOf(self.val))
    def getNum(self,*args): return(self.val)
    def setPerc(self,percentage): self.val = self._numOf(percentage)
    def setNum(self,number): self.val = number
    def set(self,user_in):
        in_type = type(user_in)
        if in_type == str and '%' in user_in: print("percentage is WIP!"); raise Exception()
        elif in_type == (int or float): self.setNum(user_in)
        else: raise TypeError("accepted formats: integer, float or string as 'x%'")
def clamp(val,minimum,maximum):
    if val < minimum: return(minimum)
    elif val > maximum: return(maximum)
    elif val >= minimum and val <= maximum: return(val)
    else: raise Exception("an error occured while clamping values")
#def percentOf(val,scaleBy): return((val/100)*scaleBy)



from _thread import start_new_thread # used to run functions in parallel
import alsaaudio as al
import time

class VolumeAssistant:
    def __init__(self):
        self.running = False
        self.volume = 10 # TODO: change this to poll the actual volume on scroll detection
        self.muted = False # TODO: change this to reflect reality
        self.previous_volume = 0
        self.device = "Master"
        self.defaultMixer = "Master"
        self.mixer = al.Mixer(self.defaultMixer)
    def start(self,pollFreq):
        def _startthread(self,pollFreq):
            self.running = True
            display = self._get_display_size()
            poll = 1/pollFreq
            while self.running == True:
                # // THREAD CODE //
                mouse = self._get_mouse_coords()
                print(self._get_percentage(mouse,display))
                time.sleep(poll)
                # // END THREAD CODE //
            print("volume assistant thread terminating...")
            print("volume assistant thread terminated.")
        start_new_thread(_startthread, (self,pollFreq))
    def stop(self):
        self.running=False
    def ismuted(self):
        _muted = self.mixer.getmute()[0]
        if _muted == 1: return(True)
        elif _muted == 0: return(False)
    def _get_percentage(self,child,parent):
        percentages = {'x': (child['x']/parent['x'])*100,
                       'y': (child['y']/parent['y'])*100}
        return(percentages)
    def _get_mouse_coords(self):
        data = display.Display().screen().root.query_pointer()._data
        return({'x': data["root_x"],
                'y': data["root_y"]})
    def _get_display_size(self,root):
        return({'x': root.winfo_screenwidth(),
                'y': root.winfo_screenheight()})
